/*******************************************************************************
Copyright 2016 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/

/** INCLUDES *******************************************************/
#include <stdint.h>
#include <stdbool.h>

#include "system.h"

#include "sreg.h"
#include "psg.h"

#include "usb.h"
#include "usb_device_midi.h"

/** VARIABLES ******************************************************/
/* Some processors have a limited range of RAM addresses where the USB module
 * is able to access.  The following section is for those devices.  This section
 * assigns the buffers that need to be used by the USB module into those
 * specific areas.
 */
#if defined(FIXED_ADDRESS_MEMORY)
    #if defined(COMPILER_MPLAB_C18)
        #pragma udata DEVICE_AUDIO_MIDI_RX_DATA_BUFFER=DEVCE_AUDIO_MIDI_RX_DATA_BUFFER_ADDRESS
            static uint8_t ReceivedDataBuffer[64];
        #pragma udata DEVICE_AUDIO_MIDI_EVENT_DATA_BUFFER=DEVCE_AUDIO_MIDI_EVENT_DATA_BUFFER_ADDRESS
            static USB_AUDIO_MIDI_EVENT_PACKET midiData;
        #pragma udata
    #elif defined(__XC8)
        static uint8_t ReceivedDataBuffer[64] @ DEVCE_AUDIO_MIDI_RX_DATA_BUFFER_ADDRESS;
        static USB_AUDIO_MIDI_EVENT_PACKET midiData @ DEVCE_AUDIO_MIDI_EVENT_DATA_BUFFER_ADDRESS;
    #endif
#else
    static uint8_t ReceivedDataBuffer[64];
    static USB_AUDIO_MIDI_EVENT_PACKET midiData;
#endif

static USB_HANDLE USBTxHandle;
static USB_HANDLE USBRxHandle;

static uint8_t pitch;
static bool sentNoteOff;

static USB_VOLATILE uint8_t msCounter;

/*********************************************************************
* Function: void APP_DeviceAudioMIDIInitialize(void);
*
* Overview: Initializes the demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceAudioMIDIInitialize()
{
    USBTxHandle = NULL;
    USBRxHandle = NULL;

    pitch = 0x3C;
    sentNoteOff = true;

    msCounter = 0;

    //enable the HID endpoint
    USBEnableEndpoint(USB_DEVICE_AUDIO_MIDI_ENDPOINT,USB_OUT_ENABLED|USB_IN_ENABLED|USB_HANDSHAKE_ENABLED|USB_DISALLOW_SETUP);

    //Re-arm the OUT endpoint for the next packet
    USBRxHandle = USBRxOnePacket(USB_DEVICE_AUDIO_MIDI_ENDPOINT,(uint8_t*)&ReceivedDataBuffer,64);
}

/*********************************************************************
* Function: void APP_DeviceAudioMIDIInitialize(void);
*
* Overview: Initializes the demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceAudioMIDISOFHandler()
{
    if(msCounter != 0)
    {
        msCounter--;
    }
}

void MIDIHandler(uint8_t* data, uint8_t sz)
{
	static uint8_t lastStatus;
	static uint8_t lastNoteNo[3];
	
	if(data[1] & 0x80){ // Status Byte
		lastStatus = data[1];
		data[1] = data[2];
		data[2] = data[3];
	}else { // may be running status
	}
	
	uint8_t ch = lastStatus & 0x0F; // Channel Number
	
	if( ((lastStatus & 0xF0) == 0x80) ||  // Note Off
		(((lastStatus & 0xF0) == 0x90) && (data[2] == 0))) // Note On and Velocity 0
	{
		// Note OFF
		
		if(ch < 3) {
			if(lastNoteNo[ch] == data[1]) SendNoteOff(ch);
		}
		else if(ch == 9){
			SendNoiseEnable(2, 0, 0);
		}
	}
	else if	(((lastStatus & 0xF0) == 0xB0) && (data[1] == 123)) // All Note Off
	{
		SendNoteOff(0);
		SendNoteOff(1);
		SendNoteOff(2);
		SendNoiseEnable(0, 0, 0);
		SendNoiseEnable(1, 0, 0);
		SendNoiseEnable(2, 0, 0);
	}
	else if((lastStatus & 0xF0) == 0x90) {
		// Note ON
		uint8_t velocity = data[2];
		
		if(ch < 3) {
			lastNoteNo[ch] = data[1];
			SendNoteOn(ch, data[1]);
		}
		else if(ch == 9) { // Noise Channel
			SendNoisePeriod(34-data[1]);
			SendNoiseEnable(2, 1, velocity/8);
		}
	}
	
	else if((lastStatus & 0xF0) == 0xB0){ 
		// Control Change
		
		if(ch < 3){
			static uint16_t envPeriod = 0;
			if(data[1] == 1) // Modulaton
			{ 
				envPeriod = (envPeriod & 0x007F) + ((uint16_t)data[2] << 7);
				SendEnvelopePeriod(envPeriod);
			}
			else if(data[1] == 11) // Expression
			{ 
				SendNoteVol(ch, data[2] >> 3);
			}
			else if(data[1] == 91)// Reverb
			{ 
			}
			else if(data[1] == 93) // Chorus
			{ 
				envPeriod = (envPeriod & 0xFF80) + (data[2]);
				SendEnvelopePeriod(envPeriod);
			}
		}
	}
	else if((lastStatus & 0xF0) == 0xC0){ // PG Change
		if(ch < 3){
			if(data[1] == 0){ // P001
				SendNoteMode(ch, 0);
			}else{ // Envelope Mode
				SendEnvelopeType(data[1]-1);
				SendNoteMode(ch, 1);
			}
		}
	}
	else if((lastStatus & 0xF0) == 0xE0){ // Pitch Bend
		if(ch < 3){
			int16_t bend = (((uint16_t)data[2]) << 7) + data[1];
			SendPitchBend(ch, bend - 8192);
		}
	}
	else if(lastStatus == 0xF0){ // System Excl.
		if(sz >= 6){
			const uint8_t GS_RESET[] = {0xF0,0x7E,0x7F,0x09,0x01,0xF7};
			uint8_t i, cond = 0;
			for(i=0; i<6; i++){
				if(data[1+i] != GS_RESET[i]) cond = 1;
			}
			if(cond == 0){
				// Reset Module
				InitPSG();
			}
		}
	}

}
/*********************************************************************
* Function: void APP_DeviceAudioMIDITasks(void);
*
* Overview: Keeps the Custom HID demo running.
*
* PreCondition: The demo should have been initialized and started via
*   the APP_DeviceAudioMIDIInitialize() and APP_DeviceAudioMIDIStart() demos
*   respectively.
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_DeviceAudioMIDITasks()
{
	/* If the device is not configured yet, or the device is suspended, then
	 * we don't need to run the demo since we can't send any data.
	 */
	if( (USBGetDeviceState() < CONFIGURED_STATE) ||
		(USBIsDeviceSuspended() == true))
	{
		return;
	}
	
	if(PORTAbits.RA3 == 0){
		// 
		SBI(PORTC, 1);
		InitPSG();
		__delay_ms(1);
		CBI(PORTC, 1);
	}

	if(!USBHandleBusy(USBRxHandle))
	{
		//We have received a MIDI packet from the host, process it and then
		//  prepare to receive the next packet

		//INSERT MIDI PROCESSING CODE HERE

		//Get ready for next packet (this will overwrite the old data)
		USBRxHandle = USBRxOnePacket(USB_DEVICE_AUDIO_MIDI_ENDPOINT, (uint8_t*)&ReceivedDataBuffer,64);
		
		MIDIHandler(ReceivedDataBuffer, 64);
		// Parse MIDI Message (ReceivedDataBuffer)
		static uint8_t cnt;
		if(cnt++ % 2){
			SBI(PORTC, 0);
		}else{
			CBI(PORTC, 0);
		}
	}

	/* If the user button is pressed... */
	if(BUTTON_IsPressed(BUTTON_DEVICE_AUDIO_MIDI) == true)
	{
		/* and we haven't sent a transmission in the past 100ms... */
		if(msCounter == 0)
		{
			/* and we have sent the NOTE_OFF for the last note... */
			if(sentNoteOff == true)
			{
				/* and we aren't currently trying to transmit data... */
				if(!USBHandleBusy(USBTxHandle))
				{
					//Then reset the 100ms counter
					msCounter = 100;

					midiData.Val = 0;   //must set all unused values to 0 so go ahead
										//  and set them all to 0

					midiData.CableNumber = 0;
					midiData.CodeIndexNumber = MIDI_CIN_NOTE_ON;
					midiData.DATA_0 = 0x90;         //Note on
					midiData.DATA_1 = pitch;         //pitch
					midiData.DATA_2 = 0x7F;  //velocity

					USBTxHandle = USBTxOnePacket(USB_DEVICE_AUDIO_MIDI_ENDPOINT,(uint8_t*)&midiData,4);

					/* we now need to send the NOTE_OFF for this note. */
					sentNoteOff = false;
				}
			}
		}
	}
	else
	{
		if(msCounter == 0)
		{
			if(sentNoteOff == false)
			{
				if(!USBHandleBusy(USBTxHandle))
				{
					//Debounce counter for 100ms
					msCounter = 100;

					midiData.Val = 0;   //must set all unused values to 0 so go ahead
										//  and set them all to 0

					midiData.CableNumber = 0;
					midiData.CodeIndexNumber = MIDI_CIN_NOTE_ON;
					midiData.DATA_0 = 0x90;     //Note off
					midiData.DATA_1 = pitch++;     //pitch
					midiData.DATA_2 = 0x00;        //velocity

					if(pitch == 0x49)
					{
						pitch = 0x3C;
					}

					USBTxHandle = USBTxOnePacket(USB_DEVICE_AUDIO_MIDI_ENDPOINT,(uint8_t*)&midiData,4);
					sentNoteOff = true;
				}
			}
		}
	}
}
